#include "config.h"
#include "version.h"
#include "tab.c"
#include "stok.c"
#include "itab.c"
#define NPATH 4096          /* max chars in a full pathname */
#define NNAME 1024          /* max chars in a path element */
#define NFMT  32            /* max number of format specifiers */
#define NSTY  32            /* max number of styles in config file */

char  EMPTY[]   = "";
char *SYS_CLSRC = "/etc/clsrc";

int      use_color = 1, use_access = 0, ls_flags = 0, mxcols = 99, gap = 1,
         mxpad = 8, width = 0, dirlabel = 0, quiet = 0,
	 exclude[32], nexclude = 0, Exclude[32], nExclude = 0,
         include[32], ninclude = 0, Include[32], nInclude = 0;
char     Fmt[]="%f", *fmt = Fmt, *ord = "tTeN", *cfg, nl[2] = "\n",
        *tfmt[10] = { "%y/%m/%d", "%H:%M:%S", "%y-%m-%d" };
time_t   age [10] = { 86400,          0,      -2147483648 }, now;
int      nage = 3;
/* "\x1b[1;34m%y\x1b[1;36m%m\x1b[1;37m%d\x1b[0m\x1b[1;34m%H\x1b[1;36m%M\x1b[1;37m%S\x1b[0m"; */
unsigned my_uid, my_gid, exit_code = 0;
itab_t  *dirtab;
char     tolow[256];        /* profiling shows this can help ~15% sometimes */
#define TOLOWER(c) (tolow[(int)(c)])

/* NOTE xterm uses color 0-15 for basic ANSI-like fg/bg, 16-231 for rgb color
       (16 + r*36 + g*6 + b), and color 232-255 for grey scale (b->w).  Below
        we accept either decimal numbers like "fg:220" or compositional phrases
	like "bg:rXgYbZ", "fg:gray-Q" for X on [0,5] and Q on [0,23]. */

char *parse_xterm256(char *buf, char *spec) {
    char *ret = buf;
    unsigned r, g, b, n, gr;

    buf += strlen(buf);
    if (sscanf(spec, "%u", &n) == 1 && spec[1] != 'x') {
	if (!(n < 256)) {
	    fprintf(stderr, "color spec n has range 0..255: \"%s\"\n", spec);
	    goto e;
	}
	strcpy(buf, spec);
	return ret;
    }
    if (sscanf(spec, "r%u,g%u,b%u", &r, &g, &b) == 3) {
	if (!(r < 6 && g < 6 && b < 6)) {
	    fprintf(stderr, "color spec r,g,b has range 0..5: \"%s\"\n", spec);
	    goto e;
	}
        sprintf(buf, "%u", 16 + 36*r + 6*g + b);
        return ret;
    }
    if (sscanf(spec, "gray-%u", &gr)==1 || sscanf(spec, "g:grey-%u", &gr)==1) {
	if (!(gr < 24)) {
	    fprintf(stderr, "gray spec has range 0..23: \"%s\"\n", spec);
	    goto e;
	}
        sprintf(buf, "%u", 232 + gr);
        return ret;
    }
    if (sscanf(spec, "0x%x", &r) == 1) { /* standard 24-bit rgb hex spec */
        sprintf(buf, "%u", 16 + 36 * (int)(6 * ((r    ) & 0xFF) / 256. + 0.5) +
		                 6 * (int)(6 * ((r>> 8) & 0xFF) / 256. + 0.5) +
				     (int)(6 * ((r>>16) & 0xFF) / 256. + 0.5) );
        return ret;
    }
e:  return (char *) 0;
}

char *attr_find(char *wd) {
  static char tmp[16];
  static struct { char *name, *code; } attrs[] = {   /* basic styles/fonts */
    {"plain"   ,  "0"},{"bold"   ,  "1"},{"italic"   ,  "3"},{"underline" , "4"},
    {"blink"   ,  "5"},{"inverse",  "7"},
/* low intensity foreground colors */
    {"black"   , "30"},{"red"    , "31"},{"green"    , "32"},{"yellow"   , "33"},
    {"blue"    , "34"},{"purple" , "35"},{"cyan"     , "36"},{"white"    , "37"},
/* high intensity foreground colors */
    {"BLACK"   , "90"},{"RED"    , "91"},{"GREEN"    , "92"},{"YELLOW"   , "93"},
    {"BLUE"    , "94"},{"PURPLE" , "95"},{"CYAN"     , "96"},{"WHITE"    , "97"},
/* low intensity background colors */
    {"on_black", "40"},{"on_red" , "41"},{"on_green" , "42"},{"on_yellow", "43"},
    {"on_blue" , "44"},{"on_cyan", "45"},{"on_purple", "46"},{"on_white" , "47"},
/* high intensity background colors */
    {"on_BLACK","100"},{"on_RED" ,"101"},{"on_GREEN" ,"102"},{"on_YELLOW","103"},
    {"on_BLUE" ,"104"},{"on_CYAN","105"},{"on_PURPLE","106"},{"on_WHITE" ,"107"}
  };
  unsigned i;
  for (i = 0; i < sizeof attrs/sizeof *attrs; i++)
    if (strcmp(attrs[i].name, wd) == 0)
      return attrs[i].code;
  if (TOLOWER(wd[0]) == 'f' && wd[1] && wd[2] && wd[3]) {
      strcpy(tmp, "38;5;");
      return parse_xterm256(tmp, wd + 3);
  }
  if (TOLOWER(wd[0]) == 'b' && wd[1] && wd[2] && wd[3]) {
      strcpy(tmp, "48;5;");
      return parse_xterm256(tmp, wd + 3);
  }
  return (char *) 0;
}

inline char *Color(char *c) { return use_color ? c : EMPTY; }

void mkattrON(char *buf, char *spec) { /* Build \e[$A;3$F;4$Bm for attr A,colr F,B */
    char *word, *s, first = 1;

    if (!spec || !*spec || !use_color) {
        *buf = '\0';
        return;
    }
    strcpy(buf, "\x1b[");
    {
        STOK_FOREACH(word, " \t", spec) {
            if (!(s = attr_find(word))) {
                stok_fix(word__tok);    /* repair writeable buffer */
                fprintf(stderr, "bad color attribute specifier \"%s\"\n", word);
                continue;
            }
            if (!first)
                strcat(buf, ";");
            strcat(buf, s);
            first = 0;
        }
    }
    strcat(buf, "m");
}

char attrOFF[] = "\x1b[0m";     /* This is common for all strings */

char *size_rnd(off_t bytes) {  /*** express file size w/4-chars */
    static const size_t K = 1UL << 10, M = 1UL << 20, G = 1UL << 30;
    static char sz[8];
    double Bytes = bytes;
    if      (bytes <= 9999   ) { sprintf(sz, "%4d", (int)bytes); return sz;  }
    else if (bytes < 99.5 * K) { sprintf(sz, "%3.2gK", Bytes/K); return sz;  }
    else if (bytes < 100  * K) { sprintf(sz, "100K");            return sz;  }
    else if (bytes < 995  * K) { sprintf(sz, "%3.3gK", Bytes/K); return sz;  }
    else if (bytes < .995 * M) { sprintf(sz, "%3.2gM", Bytes/M); return sz+1;}
    else if (bytes < 1024 * K) { sprintf(sz, "%3.2gM", Bytes/M); return sz;  }
    else if (bytes < 99.5 * M) { sprintf(sz, "%3.2gM", Bytes/M); return sz;  }
    else if (bytes < 100  * M) { sprintf(sz, "100M");            return sz;  }
    else if (bytes < 995  * M) { sprintf(sz, "%3.3gM", Bytes/M); return sz;  }
    else if (bytes < .995 * G) { sprintf(sz, "%3.2gG", Bytes/G); return sz+1;}
    else if (bytes < 1024 * M) { sprintf(sz, "%3.2gG", Bytes/G); return sz;  }
    else if (bytes < 99.5 * G) { sprintf(sz, "%3.2gG", Bytes/G); return sz;  }
    else if (bytes < 100  * G) { sprintf(sz, "100G");            return sz;  }
    else                       { sprintf(sz, "%3.3gG", Bytes/G); return sz;  }
}

char symln_ON[16], sz_bON[16], sz_kON[16], sz_mON[16], sz_gON[16];

char *size_fmt(char *sz) {               /*** colorized file size */
    char buf[32];
    int  n = strlen(sz);                   /* just to be safe - should be 4 */
    if (sz[n - 1] == 'K') {
        sprintf(buf, "%s%s%s", Color(sz_kON), sz, attrOFF);
        return strcpy(sz, buf);
    }
    if (sz[n - 1] == 'M') {
        sprintf(buf, "%s%s%s", Color(sz_mON), sz, attrOFF);
        return strcpy(sz, buf);
    }
    if (sz[n - 1] == 'G') {
        sprintf(buf, "%s%s%s", Color(sz_gON), sz, attrOFF);
        return strcpy(sz, buf);
    }
    sprintf(buf, "%s%s%s", Color(sz_bON), sz, attrOFF);
    return strcpy(sz, buf);
}

typedef struct File {	/* hopefully 128 bytes=even number of cache lines */
    struct stat    st;
    char          *usr,
                  *grp;
    signed char    type,          /* format type */
                   Rok, Wok, Xok;
    unsigned short key,           /* type sort key */
                   npfx,          /* string length of prefix */
                   nnm,           /* string length of name */
                   base,          /* offset of basename */
                   ext,           /* offset of smallest filename extension */
                   Ext;           /* offset of largest filename extension */
    char          *pfx;
    char          *name;
    struct File   *link;
} File;
typedef File *pFile;

File *path_last;

static char *path(File *f, int dirname) { /* form temporary path or dirname */
    static char buf[4096];
    int n = f->npfx;

    if (f == path_last)
        goto dn;
    path_last = f;
    if (n) {
        memcpy(buf, f->pfx, n);
        if (f->pfx[n - 1] != '/')
            buf[n++] = '/';
    }
    strcpy(buf + n, f->name);
dn: if (dirname) { /* NOTE caching only works if dirname done after path(,0)!*/
        char *p;

        for (p = buf + n + f->nnm; p > buf; p--)
            if (*p == '/')
                break;
        if (p > buf)
            *p = '\0';
    }
    return buf;
}

char permON[8][12];

char *Usr(uid_t uid);
int Gids[128], nGids;

void init_usr_grp() {
    struct group *gr;
    char **p, *me;

    my_uid = getuid();
    my_gid = getgid();
    me = Usr(my_uid);
    Gids[nGids++] = my_gid;
    while ((gr = getgrent()) && nGids < (int)(sizeof Gids/sizeof *Gids))
        for (p = gr->gr_mem; *p; p++)
            if (strcmp(me, *p) == 0) {
                Gids[nGids++] = gr->gr_gid;
                break;
            }
}

int in_grp(int g) {
    int i;

    for (i = 0; i < nGids; i++)
        if (g == Gids[i])
            return 1;
    return 0;
}

int Rok(File *f) {
    if (!f->Rok) {                      /* cache calls to access() */
        if (use_access)
            return -1+(f->Rok = 1 + (access(path(f,0), R_OK) == 0));
        else 
            return -1+(f->Rok = 1 +
                       ((my_uid == f->st.st_uid && (f->st.st_mode & S_IRUSR)) ||
                        (in_grp(f->st.st_gid)   && (f->st.st_mode & S_IRGRP)) ||
                        (f->st.st_mode & S_IROTH)));
    }
    return f->Rok - 1;
}

int Wok(File *f) {
    if (!f->Wok) {                      /* cache calls to access() */
        if (use_access)
            return -1 + (f->Wok = 1 + (access(path(f,0), W_OK) == 0));
        else 
            return -1+(f->Wok = 1 +
                       ((my_uid == f->st.st_uid && (f->st.st_mode & S_IWUSR)) ||
                        (in_grp(f->st.st_gid)   && (f->st.st_mode & S_IWGRP)) ||
                        (f->st.st_mode & S_IWOTH)));
    }
    return f->Wok - 1;
}

int Xok(File *f) {
    if (!f->Xok) {                      /* cache calls to access() */
        if (use_access)
            return -1+(f->Xok = 1 + (access(path(f,0), X_OK) == 0));
        else 
            return -1+(f->Xok = 1 +
                       ((my_uid == f->st.st_uid && (f->st.st_mode & S_IXUSR)) ||
                        (in_grp(f->st.st_gid)   && (f->st.st_mode & S_IXGRP)) ||
                        (f->st.st_mode & S_IXOTH)));
    }
    return f->Xok - 1;
}

char *perm2rwx(int m) {
    static char  buf[16];
    static char *rwx[]={"---", "--x", "-w-", "-wx", "r--", "r-x", "rw-", "rwx"};

    sprintf(buf, "%s%s%s", rwx[m >> 6 & 7], rwx[m >> 3& 7], rwx[m & 7]);
    if (m & 04000)
	buf[2] = 's';
    if (m & 02000)
	buf[5] = 's';
    if (m & 01000)
	buf[8] = 't';
    return buf;
}

int isOther      (File *f, void *a) { return 1; }
int isDot        (File *f, void *a) { return f->name[f->base] == '.'; }
int isSymLink    (File *f, void *a) { return S_ISLNK(f->st.st_mode); }
int isDirectory  (File *f, void *a) { return S_ISDIR(f->st.st_mode); }
int isSocket     (File *f, void *a) { return S_ISSOCK(f->st.st_mode); }
int isBlockDev   (File *f, void *a) { return S_ISBLK(f->st.st_mode); }
int isCharDev    (File *f, void *a) { return S_ISCHR(f->st.st_mode); }
int isExecutable (File *f, void *a) { return Xok(f); }
int isBrokenLink (File *f, void *a) {
    struct stat st;
    return isSymLink(f, a) && stat(path(f,0), &st) == -1;
}

struct {
    char name[256];
    int (*test)(File *f, void *aux);
    void *aux;
    int   slot;
} AvailTests[128] = {
   { "other",       isOther      },
   { "symlink",     isSymLink    },
   { "directory",   isDirectory  },
   { "socket",      isSocket     },
   { "blockdev",    isBlockDev   },
   { "chardev",     isCharDev    },
   { "executable",  isExecutable },
   { "brokenlink",  isBrokenLink },
   { "dot",         isDot        }
};
int nAvailTests = 9;                    /* NOTE important to keep in sync! */

typedef struct {
    char attrON[32];
    int  key;
    int (*test)(File *f, void *aux);
    void *aux;
} Type;
Type Types[128], KeyTypes[128];
unsigned nTypes = 0;

int cmp_int(const int a, const int b) {
    if (a < b)
        return -1;
    if (a > b)
        return +1;
    return 0;
}

int cmp_off(const off_t a, const off_t b) {
    if (a < b)
        return -1;
    if (a > b)
        return +1;
    return 0;
}

int TypeCmp(const void *a, const void *b) {
    return cmp_int( ((Type *)a)->key, ((Type *)b)->key);
}

int typeBroken = -1;    /* need type number for broken in File_make() */
int typeDir    = -1;    /* need type number for dirs in lsdir() */

int test_chlist(File *f, void *cc) {
    char *p, *CC = (char *)cc;

    for (p = f->name + f->base; *p; p++)
	if (!CC[(int)*p])
	    return 0;
    return 1;
}

int add_chlist(char *name, char *chlist) { /* add to AvailTests */
    char *p;
    if (nAvailTests == sizeof AvailTests / sizeof *AvailTests)
        return fprintf(stderr, "too many types at: \"%s\"\n", name), 1;
    AvailTests[nAvailTests].test = test_chlist;
    strcpy(AvailTests[nAvailTests].name, name);
    if (!(AvailTests[nAvailTests].aux = calloc(256,1)))
        return fprintf(stderr, "out of memory for: \"%s\"\n", name), 1;
    for (p = chlist; *p; p++)
	((char*)AvailTests[nAvailTests].aux)[(int)*p] = 1;
    nAvailTests++;
    return 0;
}

int test_Suffix(File *f, void *sfx) {
    char **v = (char **) sfx, *end = f->name + f->nnm - 1, *p, *q;

    for (/**/; *v; v++) {
        for (p = *v, q = end; *p && *q && *q == *p; p++, q--)
            /**/;
        if (!*p)        /* consumed all of this sfx */
            return 1;
    }
    return 0;
}

int test_suffix(File *f, void *sfx) {   /* case-insensitive test_Suffix */
    char **v = (char **) sfx, *end = f->name + f->nnm - 1, *p, *q;

    for (/**/; *v; v++) {
        for (p = *v, q = end; *p && *q && TOLOWER(*q) == *p; p++, q--)
            /**/;
        if (!*p)        /* consumed all of this sfx */
            return 1;
    }
    return 0;
}

int add_suffix(char *name, char *kind, char *sfxlist) { /* add to AvailTests */
    char **compiled, *p, *q, *sfx;
    int    n = strlen(sfxlist), nsfx = 0;

    if (nAvailTests == sizeof AvailTests / sizeof *AvailTests)
        return fprintf(stderr, "too many types at: \"%s\"\n", name), 1;
    if (!(compiled = (char **) malloc((n + 1) + 128)))
        return fprintf(stderr, "out of memory\n"), 2;
    if (kind[1] == 'S')
        AvailTests[nAvailTests].test = test_Suffix;
    else if (kind[1] == 's') {
        for (p = sfxlist; *p; p++)
            *p = TOLOWER(*p);
        AvailTests[nAvailTests].test = test_suffix;
    } else
        return fprintf(stderr, "bad assignment operator for: \"%s\"\n", name),3;
    { STOK_FOREACH(sfx, " \t", sfxlist) {
        compiled[nsfx] = strdup(sfx);
        for (p = sfx + strlen(sfx) - 1, q = compiled[nsfx]; p >= sfx; p--, q++)
            *q = *p;
        nsfx++;
    }}
    compiled[nsfx] = NULL;
    strcpy(AvailTests[nAvailTests].name, name);
    AvailTests[nAvailTests].aux = compiled;
    nAvailTests++;
    return 0;
}

int test_pattern(File *f, void *rx) {
    regmatch_t pm[128];
    return regexec((regex_t *)rx, f->name, sizeof pm/sizeof *pm, pm, 0) == 0;
}

int add_pattern(char *name, char *kind, char *pattern) { /* add to AvailTests */
    regex_t *compiled;
    int      flags = REG_EXTENDED;

    if (nAvailTests == sizeof AvailTests / sizeof *AvailTests)
        return fprintf(stderr, "too many types at: \"%s\"\n", name), 1;
    if (!(compiled = (regex_t *) malloc(sizeof *compiled)))
        return fprintf(stderr, "out of memory\n"), 2;
    if (kind[1] == 'r')
        flags |= REG_ICASE;
    else if (kind[1] != 'R')
        return fprintf(stderr, "bad assignment operator for: \"%s\"\n", name),3;
    if (regcomp(compiled, pattern, flags) != 0) {
        free((void*) compiled);
        return fprintf(stderr, "bad regex: \"%s\"\n", pattern), 4;
    }
    strcpy(AvailTests[nAvailTests].name, name);
    AvailTests[nAvailTests].test = test_pattern;
    AvailTests[nAvailTests].aux = compiled;
    nAvailTests++;
    return 0;
}

int add_test(char *name, int key, char *color) {        /* add to Types[] */
    int i;

    for (i = 0; i < nAvailTests; i++)
        if (strcmp(name, AvailTests[i].name) == 0)
            break;
    if (i == nAvailTests)
        return fprintf(stderr, "no such type test: \"%s\"\n", name), 1;
    if (nTypes == sizeof Types/sizeof *Types)
        return fprintf(stderr, "too many file types at: \"%s\"\n", name), 2;
    mkattrON(Types[nTypes].attrON, color);
    Types[nTypes].key  = key;
    Types[nTypes].test = AvailTests[i].test;
    Types[nTypes].aux  = AvailTests[i].aux;
    AvailTests[i].slot = nTypes;
    if (Types[nTypes].test == isBrokenLink)
        typeBroken = nTypes;
    if (Types[nTypes].test == isDirectory)
        typeDir = nTypes;
    nTypes++;
    return 0;
}

char  *v4style_buf[NSTY * NSTY];  /* 2D array uses storage map.  Cannot use! */
char **v4style[NSTY];          
int    n4style[NSTY];          
int    nstyles = 0;               /* style 0=reserved. element 0=style name */

void init_style() {
    int i;
    for (i = 0; i < NSTY; i++)
        v4style[i] = (char**)((char*) v4style_buf + (NSTY * i * sizeof(char*)));
}

int find_style(char *name) {
    int i;

    if (nstyles == 0) {
        nstyles = 1;
        init_style();
    }
    for (i = 1; i < nstyles; i++)
        if (strcmp(name, v4style[i][0]) == 0)
            break;
    if (i == nstyles)
        return 0;
    return i;
}

int add_style(char **w, int n) {   /* split, alloc copy, add to style table */
    unsigned i;
    char    *arg;

    if (n < 3)                          /* ignore only style and name */
        return fprintf(stderr, "No style definition for %s\n", w[1]), 1;
    if (!(i = find_style(w[1])))                /* new style */
        if ((i = nstyles++) > sizeof n4style/sizeof n4style[0] - 1)
            return fprintf(stderr, "Too many styles\n"), 0;
    if (!n4style[i]) {
        v4style[i][0] = strdup(w[1]);
        n4style[i] = 1;
    }
    {STOK_FOREACH(arg, "  \t", w[2]) {  /* parse w[2], onto v4style[i][] */
        if (arg[0] == '-' && arg[1] == 'f')
            stok_fix(arg__tok);
        v4style[i][n4style[i]++] = strdup(arg);
        if (arg[0] == '-' && arg[1] == 'f')
            break;
    }}
    v4style[i][n4style[i]] = NULL;
    return 0;
}

int parse_config(char *buf) {
    unsigned i, sort_key, n, errs = 0;
    char *line, *w[3];

    STOK_FOREACH(line, "\n\n", buf) {
        if (line[0] == '#')                           continue; /* comment line */
        if (!*(line + strspn(line, " \t")))           continue; /* blank line */
        if ((n = stok_splitN(line, " \t", w, 3)) < 2) continue; /* no args! */
        if (strcmp(w[0], "color") == 0) {
            if (n < 3)
                errs++;
            else if (!strcmp(w[1], "symlink")) mkattrON(symln_ON,  w[2]);
            else if (!strcmp(w[1], "sizeB" ))  mkattrON(sz_bON,    w[2]);
            else if (!strcmp(w[1], "sizeKB"))  mkattrON(sz_kON,    w[2]);
            else if (!strcmp(w[1], "sizeMB"))  mkattrON(sz_mON,    w[2]);
            else if (!strcmp(w[1], "sizeGB"))  mkattrON(sz_gON,    w[2]);
            else if (!strcmp(w[1], "perm0"))   mkattrON(permON[0], w[2]);
            else if (!strcmp(w[1], "perm1"))   mkattrON(permON[1], w[2]);
            else if (!strcmp(w[1], "perm2"))   mkattrON(permON[2], w[2]);
            else if (!strcmp(w[1], "perm3"))   mkattrON(permON[3], w[2]);
            else if (!strcmp(w[1], "perm4"))   mkattrON(permON[4], w[2]);
            else if (!strcmp(w[1], "perm5"))   mkattrON(permON[5], w[2]);
            else if (!strcmp(w[1], "perm6"))   mkattrON(permON[6], w[2]);
            else if (!strcmp(w[1], "perm7"))   mkattrON(permON[7], w[2]);
            else
                errs++;
        } else if (strcmp(w[0], "style") == 0)
            errs += add_style(w, n);
        else if (w[1][0] == '=' && TOLOWER(w[1][1]) == 'c') /* chlist def */
            errs += add_chlist(w[0], w[2]);
        else if (w[1][0] == '=' && TOLOWER(w[1][1]) == 'r') /* pattern def */
            errs += add_pattern(w[0], w[1], w[2]);
        else if (w[1][0] == '=' && TOLOWER(w[1][1]) == 's') /* suffix def */
            errs += add_suffix(w[0], w[1], w[2]);
        else if ((sort_key = atoi(w[1]))) {                 /* type def */
            char *p;
            if (w[2] && (p = strchr(w[2], '#')))
                *p = '\0';
            errs += add_test(w[0], sort_key, w[2] ? w[2] : EMPTY);
        } else
            errs++;
    }
    if (Types[nTypes - 1].test != isOther) {
        fprintf(stderr, "Type tests must end in built-in \"other\" type\n");
        errs++;
    }
    if (typeBroken == -1)                       /* use Other when Broken */
        typeBroken = nTypes - 1;                /* is not available */
    if (typeDir == -1)
        typeDir = nTypes - 1;
    for (i = 0; i < nTypes; i++)
        KeyTypes[i] = Types[i];
    qsort(KeyTypes, nTypes - 1 /* leave other */, sizeof KeyTypes[0], TypeCmp);
    return errs;
}

char *Usr(uid_t uid) {
    static uid_t   uids[256], n, i;
    static char    name[256][8];
    struct passwd *pw;

    for (i = 0; i < n; i++)
        if (uid == uids[i])
            break;
    if (i < n)
        return name[i];
    if (n == 256) {
        fprintf(stderr, "too many distinct uids\n");
        exit(1);
    }
    uids[n] = uid;
    if ((pw = getpwuid(uid)))
        strcpy(name[n], pw->pw_name);
    else
        sprintf(name[n], "%u", uid);        /* fall back to numeric id */
    n++;
    return name[n - 1];
}

char *Grp(gid_t gid) {
    static gid_t   gids[256], n, i;
    static char    name[256][8];
    struct group  *gr;

    for (i = 0; i < n; i++)
        if (gid == gids[i])
            break;
    if (i < n)
        return name[i];
    if (n == 256) {
        fprintf(stderr, "too many distinct gids\n");
        exit(1);
    }
    gids[n] = gid;
    if ((gr = getgrgid(gid)))
        strcpy(name[n], gr->gr_name);
    else
        sprintf(name[n], "%u", gid);        /* fall back to numeric id */
    n++;
    return name[n - 1];
}

#define CLS_TYPE      0x001     /* need to type the file for format */
#define CLS_KEY       0x002     /* need to type the file for sorting */
#define CLS_DEREF     0x004
#define CLS_NEED_TGT  0x008
#define CLS_DEREF_TGT 0x010
#define CLS_TYPE_TGT  0x020
#define CLS_USER      0x040
#define CLS_GROUP     0x080
#define CLS_RECURSING 0x100
#define CLS_NODIR     0x200     /* elide dirs frome av[] */
#define CLS_EXPLICIT  0x400

char *Pfx(char *pfx) {
    if (pfx[0] == '.' && !pfx[1])
        pfx++;
    while (pfx[0] == '.' && pfx[1] == '/')      /* remove unecessary pfx */
        pfx += 2;
    return pfx;
}

/* NOTE: Assume pfx, name point to untouched memory for lifetime of object. */
File *File_make(File *f, int flags, char *pfx, int npfx, char *name, int nnm) {
    unsigned i;
    char    *p;

    f->pfx  = pfx ? pfx : EMPTY;
    f->npfx = npfx;
    f->name = name ? name : EMPTY;
    f->nnm  = strlen(name);     /*XXX many sys could get from d_namlen,etc... */
    f->base = (p = strrchr(name, '/')) ? p - name + 1 : 0;
    f->ext  = (p = strrchr(name, '.')) ? p - name + 1 : 0;
    f->Ext  = (p = strchr (name, '.')) ? p - name + 1 : 0;
    f->link = NULL;
    f->Rok = f->Wok = f->Xok = 0;
    /* Even CLS_DEREF lstat()s to embellish "from symlink" files, so the logic
       here is to first lstat() and then conditionally stat() */
    if (lstat(path(f, 0), &f->st) == -1) {
	memset((char *)&f->st, 0, sizeof f->st);
	f->type = typeBroken;		/* vanished since readdir | bogus tgt */
	if (flags & CLS_RECURSING)	/* only link target is not-existent */
	    goto dn;			/* skip setting type based on st */
	else {
	    fprintf(stderr, "%s: %s\n", path(f, 0), strerror(errno));
	    exit_code = 1;
	    f->key = 0;
	    return NULL;
	}
    }
    if (flags & CLS_DEREF) {
	if (S_ISLNK(f->st.st_mode))	/* remember if symlink for italics */
	    f->link = (File *) 1;	/* pointer useless in DEREF mode */
        if (stat(path(f, 0), &f->st) == -1) {
            memset((char *)&f->st, 0, sizeof f->st);
            f->type = typeBroken;
            goto dn;			/* skip setting type based on st */
        }
    }
    if (flags & CLS_TYPE) {
        for (i = 0; i < nTypes - 1; i++)
            if (Types[i].test(f, Types[i].aux))
                break;
        f->type = i;    /* Types[nTypes - 1]='other' which is like /bin/true */
    }
dn: if (flags & CLS_USER)
        f->usr = Usr(f->st.st_uid);
    if (flags & CLS_GROUP)
        f->grp = Grp(f->st.st_gid);
    if (flags & CLS_KEY) {
        for (i = 0; i < nTypes - 1; i++)
            if (KeyTypes[i].test(f, KeyTypes[i].aux))
                break;
        f->key = KeyTypes[i].key;
    }
    if (isSymLink(f, NULL) && (flags&CLS_NEED_TGT) && !(flags&CLS_RECURSING)) {
	int m;
	f->link = (File *) malloc(sizeof(File));
	f->link->name = (char *) malloc(1024);
	flags &= ~CLS_DEREF;
	if (flags & CLS_DEREF_TGT)
	    flags |= CLS_DEREF;
	flags |= CLS_RECURSING;
	path_last = NULL;
        if ((m = readlink(path(f, 0), f->link->name, 1024)) == -1) {
            fprintf(stderr, "readlink(\"%s\"): %s\n",path(f,0),strerror(errno));
	    free(f->link->name);
	    free(f->link);
	    f->link = NULL;
            return f;
        }
        f->link->name[m] = '\0';
        f->link->name = realloc(f->link->name, m + 1);
        if (f->link->name[0] == '/')              /* absolute symlink */
            File_make(f->link, flags, NULL, 0, f->link->name, m);
        else {		/* XXX check relative link -- current pfx + name */
            File_make(f->link, flags, pfx, npfx, f->link->name, m);
        }
	if (flags & CLS_DEREF_TGT)
	    f->key = f->link->key;
    }
    return f;
}

int parse_typelist(char *spec, int *list, int *n, int mode) {
    unsigned i, N = 0, errs = 0;
    char *type;

    STOK_FOREACH(type, ", ", spec) {
        for (i = 0; i < nTypes; i++)
            if (strcmp(type, AvailTests[i].name) == 0) {
                list[N] = mode ? AvailTests[i].slot : Types[AvailTests[i].slot].key;
                break;
            }
        if (i == nTypes) {
	    if (!quiet) {
		fprintf(stderr, "no such type \"%s\"\n", type);
		errs++;
	    }
        } else
            N++;
    }
    *n = (int) N;
    return errs;
}

time_t *Mtime(File *f) {
    if (f->st.st_ctime > f->st.st_mtime)
        return &f->st.st_ctime;
    return &f->st.st_mtime;
}

int  fmt_filename (char *a, int n, File *f) {
    return snprintf(a, n, "%s%s%s%s",
	     ls_flags&CLS_DEREF && f->link ? Color(symln_ON) : "",
	     Color(Types[f->type].attrON), f->name, attrOFF);
}
int  fmt_basename (char *a, int n, File *f) {
    return snprintf(a, n, "%s%s%s%s",
	     ls_flags&CLS_DEREF && f->link ? Color(symln_ON) : "",
	     Color(Types[f->type].attrON), f->name + f->base, attrOFF);
}
int  fmt_blocks   (char *a, int n, File *f) {
    return snprintf(a, n, CLS_BLOCKS_FMT, f->st.st_blocks);
}
int  fmt_i_node   (char *a, int n, File *f) {
    return snprintf(a, n, "%lu", (unsigned long)(f->st.st_ino));
}
int  fmt_ln_cnt   (char *a, int n, File *f) {
    return snprintf(a, n, "%ld", f->st.st_nlink);
}
int  fmt_size     (char *a, int n, File *f) {
    return snprintf(a, n, CLS_SIZE_FMT, f->st.st_size);
}
int  fmt_devminor (char *a, int n, File *f) {
    return snprintf(a, n, "%d", (int)(f->st.st_rdev & 0xFF));
}
int  fmt_devmajor (char *a, int n, File *f) {
    return snprintf(a, n, "%d", (int)(f->st.st_rdev >> 8));
}
int  fmt_usr_id   (char *a, int n, File *f) {
    return snprintf(a, n, "%d", f->st.st_uid);
}
int  fmt_grp_id   (char *a, int n, File *f) {
    return snprintf(a, n, "%d", f->st.st_gid);
}
int  fmt_time     (char *a, int n, time_t *t) {
    struct tm *tm = localtime(t);
    int i;
    for (i = 0; i < nage; i++)
	if (now - *t >= age[i])
	    return strftime(a, n, tfmt[i], tm);
    return strftime(a, n, tfmt[nage - 1], tm);
}
int  fmt_atime    (char *a, int n, File *f) {
    return fmt_time(a, n, &f->st.st_atime);
}
int  fmt_mtime    (char *a, int n, File *f) {
    return fmt_time(a, n, &f->st.st_mtime);
}
int  fmt_ctime    (char *a, int n, File *f) {
    return fmt_time(a, n, &f->st.st_ctime);
}
int  fmt_Mtime    (char *a, int n, File *f) {
    return fmt_time(a, n, Mtime(f));
}
int  fmt_usr_nm   (char *a, int n, File *f) {
    return strlen(strncpy(a, f->usr ? f->usr : EMPTY, n));
}
int  fmt_grp_nm   (char *a, int n, File *f) {
    return strlen(strncpy(a, f->grp ? f->grp : EMPTY, n));
}
int  fmt_perm     (char *a, int n, File *f) {
    return strlen(strncpy(a, perm2rwx(f->st.st_mode & 4095), n));
}
int  fmt_operm    (char *a, int n, File *f) {
    int m = f->st.st_mode & 4095;
    int p = Rok(f)<<2 | Wok(f)<<1 | Xok(f); /* color by perms of THIS process */
    return snprintf(a, n, "%s%04o%s", Color(permON[p]), m, Color(attrOFF));
}
int  fmt_ord_type (char *a, int n, File *f) {
    return snprintf(a, n, "%d", f->key);
}
int  fmt_fmt_type (char *a, int n, File *f) {
    return snprintf(a, n, "%d", f->type);
}
int  fmt_kindcode (char *a, int n, File *f) {
    if      S_ISLNK(f->st.st_mode)  *a = 'l';
    else if S_ISDIR(f->st.st_mode)  *a = 'd';
    else if S_ISSOCK(f->st.st_mode) *a = 's';
    else if S_ISBLK(f->st.st_mode)  *a = 'b';
    else if S_ISCHR(f->st.st_mode)  *a = 'c';
    else                            *a = '-';
    return 1;
}
int  fmt_readlink (char *a, int n, File *f) {
    char tgt[1024];
    int  m;
    if (isSymLink(f, NULL)) {
	if (f->link > (File *) 1)
	    return snprintf(a, n, "->%s", f->link->name);
	else {
	    path_last = NULL;
	    if ((m = readlink(path(f, 0), tgt, sizeof tgt)) == -1) {
		fprintf(stderr, "readlink(\"%s\"): %s\n", path(f,0), strerror(errno));
                goto x;
	    }
	    tgt[m] = '\0';
	    return snprintf(a, n, "->%s", tgt);
	}
    }
x:  *a = '\0';
    return 0;
}
int  fmt_link_tgt (char *a, int n, File *f) {
    if (f->link > (File *)1)
	return snprintf(a, n, "->%s%s%s", Color(Types[f->link->type].attrON),
                        f->link->name, attrOFF);
    *a = '\0';
    return 0;
}
int  fmt_sz_or_dev(char *a, int n, File *f) {
    if (S_ISBLK(f->st.st_mode) || S_ISCHR(f->st.st_mode))
        return snprintf(a, n, "%s%04X%s", Color("[4m"),
                        (unsigned)(f->st.st_rdev), Color(attrOFF));
    return snprintf(a, n, size_fmt(size_rnd(f->st.st_size)));
}

struct {
    char    tag,                        /* tag for for %-specifier parsing */
            pad,                        /* padding: 0=none, 1=left, 2=right */
           *desc;                       /* description for usage message */
    int   (*fmt)(char *, int, File *);  /* like snprintf() returning used len */
} AvailFmts[] = {
    { 'f', 0, "file name"   , fmt_filename  },
    { 'F', 0, "base name"   , fmt_basename  },
    { 'b', 1, "file blocks" , fmt_blocks    },
    { 'i', 1, "i-node num"  , fmt_i_node    },
    { 'n', 1, "link count"  , fmt_ln_cnt    },
    { 'S', 1, "size(bytes)" , fmt_size      },
    { 'd', 0, "minor devno" , fmt_devminor  },
    { 'D', 0, "major devno" , fmt_devmajor  },
    { 'u', 1, "user id"     , fmt_usr_id    },
    { 'g', 1, "group id"    , fmt_grp_id    },
    { 'a', 2, "access time" , fmt_atime     },
    { 'm', 2, "modify time" , fmt_mtime     },
    { 'c', 2, "create time" , fmt_ctime     },
    { 'M', 2, "max(ctm,mtm)", fmt_Mtime     },
    { 'U', 2, "user name"   , fmt_usr_nm    },
    { 'G', 2, "group name"  , fmt_grp_nm    },
    { 'p', 0, "permission"  , fmt_perm      },
    { 'P', 0, "octal perms" , fmt_operm     },
    { 'k', 0, "ls-kindcode" , fmt_kindcode  },
    { 't', 1, "order type"  , fmt_ord_type  },
    { 'T', 1, "format type" , fmt_fmt_type  },
    { 'r', 0, "readlink"    , fmt_readlink  },
    { 'R', 0, "color lnktgt", fmt_link_tgt  },
    { 's', 0, "hr size|dno" , fmt_sz_or_dev }
};

char  *fmtstr[NFMT];
int  (*fmts  [NFMT])(char *, int, File *);
int    nfmts, fmt_pad[NFMT];

int parse_fmtspec(char *a) {    /*** convert %..? in 'a' -> %..s, accum fmts */
    unsigned i;
    nfmts = 0;
    fmtstr[nfmts] = a;
    for (; *a; a++) {
        if (*a != '%')                          /* XXX collect non-spec text */
            continue;
        if (*++a == '%')                        /* skip double % */
            continue;
        a += strspn(a, "#0123456789+-. ");      /* skip printf internals */
        for (i = 0; i < sizeof AvailFmts/sizeof *AvailFmts; i++)
            if (*a == AvailFmts[i].tag) {
                *a = '\0';                      /* will later add 's' char */
                if (nfmts == NFMT) continue;
                fmt_pad[nfmts] = AvailFmts[i].pad;
                fmts[nfmts++]  = AvailFmts[i].fmt;
                fmtstr[nfmts]  = a + 1;
                if (AvailFmts[i].fmt == fmt_usr_nm)   ls_flags |= CLS_USER;
                if (AvailFmts[i].fmt == fmt_grp_nm)   ls_flags |= CLS_GROUP;
                if (AvailFmts[i].fmt == fmt_link_tgt) ls_flags |= CLS_NEED_TGT;
                break;
            }
        if (i == sizeof AvailFmts/sizeof *AvailFmts)
            return fprintf(stderr, "unknown specifier \"%%%c\"\n", *a), 0;
    }
    return 0;
}

void format(File **f, int n, char **S, int *W) { /** drive formatting */
    int    i, j, k, w, w_max[NFMT] = { 0 };
    char **field = (char **)malloc(nfmts * n * sizeof *field), buf[2048], *a;

    for (i = 0; i < n; i++) {
        for (j = 0; j < nfmts; j++) {
            w = fmts[j](buf, sizeof buf, f[i]);	/* format field j of file i */
            field[nfmts*i + j] = malloc(w + 3); /* +2 for len, +1 for NUL */
            memcpy(field[nfmts*i + j] + 2, buf, w + 1);
            if (w > w_max[j])			/* track max width */
                w_max[j] = w;
            *(short*)field[nfmts*i + j] = w;
        }
    }
    for (i = 0; i < n; i++) {                   /* padding/collating phase */
        a = buf;
        for (j = 0; j < nfmts; j++) {
            int len = *(short *)(field[nfmts*i + j]), pad = w_max[j] - len;
            if (fmt_pad[j] == 1) for (k = 0; k < pad; k++) *a++ = ' ';
            memcpy(a, field[nfmts*i + j] + 2, len); a += len;
            if (fmt_pad[j] == 2) for (k = 0; k < pad; k++) *a++ = ' ';
            if (j < nfmts - 1 && fmts[j] != fmt_kindcode &&
                fmts[j+1] != fmt_link_tgt && fmts[j+1] != fmt_readlink)
                *a++ = ' ';
            *a = 0;
            free(field[nfmts*i + j]);
        }
        S[i] = strdup(buf);
        W[i] = tab_printed_width(buf);
    }
    free(field);
}

/* strcmp that compares "parallel" numeric substrings as numbers.  Note that
   all we really must do is (for differing strings) back up the scan from the
   first differing byte to the beginning of the number.  Indeed we only even
   need do this when the differing bytes are themselves both numeric. */

int strcmpN(const char *a, const char *b) {
    unsigned long A, B;
    const char   *p = a;

    while (*a && *b && *a == *b)
        a++, b++;
    if (!*a && !*b)
        return 0;
    if (!(isdigit(*a) && isdigit(*b)))
        return cmp_int((int)*a, (int)*b);
    for (B = 0; a > p && isdigit(a[-1]); a--)   /* backup a */
        B++;                                    /* remember how much back up */
    A = strtoul(a, NULL, 10);
    b = b - B;                                  /* backup b */
    B = strtoul(b, NULL, 10);
    return cmp_int(A, B);
}

#define CMP_DEF(__NAME, __CMP, __a, __b)        \
int __NAME(const void *A, const void *B) {      \
        File *a = *(pFile *) A;                 \
        File *b = *(pFile *) B;                 \
        return __CMP((__a), (__b));             \
}
CMP_DEF(cmp_name      , strcmp , a->name             , b->name             )
CMP_DEF(cmp_basename  , strcmp , a->name + a->base   , b->name + b->base   )
CMP_DEF(cmp_blocks    , cmp_int, a->st.st_blocks     , b->st.st_blocks     )
CMP_DEF(cmp_inode     , cmp_int, a->st.st_ino        , b->st.st_ino        )
CMP_DEF(cmp_linkcount , cmp_int, a->st.st_nlink      , b->st.st_nlink      )
CMP_DEF(cmp_size      , cmp_off, a->st.st_size       , b->st.st_size       )
CMP_DEF(cmp_minordevno, cmp_int, a->st.st_rdev & 0xFF, b->st.st_rdev & 0xFF)
CMP_DEF(cmp_majordevno, cmp_int, a->st.st_rdev >> 8  , b->st.st_rdev >> 8  )
CMP_DEF(cmp_userid    , cmp_int, a->st.st_uid        , b->st.st_uid        )
CMP_DEF(cmp_groupid   , cmp_int, a->st.st_gid        , b->st.st_gid        )
CMP_DEF(cmp_atime     , cmp_int, a->st.st_atime      , b->st.st_atime      )
CMP_DEF(cmp_mtime     , cmp_int, a->st.st_mtime      , b->st.st_mtime      )
CMP_DEF(cmp_ctime     , cmp_int, a->st.st_ctime      , b->st.st_ctime      )
CMP_DEF(cmp_Mtime     , cmp_int, *Mtime(a)           , *Mtime(b)           )
CMP_DEF(cmp_username  , strcmp , a->usr              , b->usr              )
CMP_DEF(cmp_groupname , strcmp , a->grp              , b->grp              )
CMP_DEF(cmp_permission, cmp_int, a->st.st_mode & 4095, b->st.st_mode & 4095)
CMP_DEF(cmp_type_key  , cmp_int, a->key              , b->key              )
CMP_DEF(cmp_type_no   , cmp_int,
    ((ls_flags&CLS_TYPE_TGT)&& a->link > (File *)1 ? a->link->type : a->type),
    ((ls_flags&CLS_TYPE_TGT)&& b->link > (File *)1 ? b->link->type : b->type))
CMP_DEF(cmp_extension , strcmpN, a->name + a->ext    , b->name + b->ext    )
CMP_DEF(cmp_Extension , strcmpN, a->name + a->Ext    , b->name + b->Ext    )
CMP_DEF(cmp_namelen   , cmp_int, a->nnm              , b->nnm              )
CMP_DEF(cmp_numeric   , strcmpN, a->name             , b->name             )

struct {
    char tag,
        *desc;
    int (*cmp)(const void *A, const void *B);
} AvailCmps[] = {
    { 'f', "file name",                 cmp_name        },
    { 'F', "base name",                 cmp_basename    },
    { 'N', "numeric file names",        cmp_numeric     },
    { 'b', "file blocks",               cmp_blocks      },
    { 'i', "i-node number",             cmp_inode       },
    { 'n', "link count",                cmp_linkcount   },
    { 's', "size",                      cmp_size        },
    { 'd', "minor dev num",             cmp_minordevno  },
    { 'D', "major dev num",             cmp_majordevno  },
    { 'u', "user id",                   cmp_userid      },
    { 'g', "group id",                  cmp_groupid     },
    { 'a', "access time",               cmp_atime       },
    { 'm', "modify time",               cmp_mtime       },
    { 'c', "create time",               cmp_ctime       },
    { 'M', "max(ctime,mtime)",          cmp_Mtime       },
    { 'U', "user name",                 cmp_username    },
    { 'G', "group name",                cmp_groupname   },
    { 'p', "permission",                cmp_permission  },
    { 't', "order type number",         cmp_type_key    },
    { 'T', "format type number",        cmp_type_no     },
    { 'e', "smallest extension",        cmp_extension   },
    { 'E', "largest extension",         cmp_Extension   },
    { 'L', "filename length",           cmp_namelen     }
};

struct  {
    int    n;                                      /* number of comparers */
    int    sgn[8];                                 /* comparison directions */
    int  (*cmp[8])(const void *, const void*);     /* compare functions */
} cmps;

int parse_cmpspec(char *a) {
    char    *p;
    unsigned i;
    int      sgn = +1;

    cmps.n = 0;
    if (strcmp(a, "-") == 0)
        return 0;
    for (p = a; *p; p++) {
        if (*p == '-') {
            sgn = -1;
            continue;
        }
        for (i = 0; i < sizeof AvailCmps/sizeof *AvailCmps; i++)
            if (*p == AvailCmps[i].tag)
                break;
        if (i == sizeof AvailCmps/sizeof *AvailCmps)
            return fprintf(stderr, "unknown sort key specifier '%c'\n", *p),1;
        cmps.cmp[cmps.n] = AvailCmps[i].cmp;
        if (AvailCmps[i].cmp == cmp_username)
            ls_flags |= CLS_USER;
        if (AvailCmps[i].cmp == cmp_groupname)
            ls_flags |= CLS_GROUP;
        cmps.sgn[cmps.n] = sgn;
        cmps.n++;
        sgn = +1;
    }
    return 0;
}

inline int MultiLevelCmp(const void *a, const void *b) {
    int i, val;
    for (i = 0; i < cmps.n; i++)
        if ((val = cmps.cmp[i](a, b)))
            return cmps.sgn[i] * val;
    return 0;
}

int int_in(int key, int *list, int n) {
    int i;
    for (i = 0; i < n; i++)
        if (list[i] == key)
            return 1;
    return 0;
}

/* Profiling shows a lot of extra time is in sort.  So, we optimize that */
#define S_IDX_T  unsigned
#define S_VAL_T  pFile
#define S_CMP    MultiLevelCmp
#define S_SORT   File_sort
#define S_(x)    File_sort_ ## x
#define CUSTOM_SORT
#include "sort.c"

void ls(char *pfx, char **args, int n, int flags, char *isdir) { /* 1-DIR LS */
    File  *f, *files = (File *) malloc(n * sizeof(File)), **list;
    char **p, **strs;
    int    npfx, nfiles, i, *lens, *ws, nrow, ncol;

    pfx  = Pfx(pfx);
    npfx = strlen(pfx);
    for (i = 0, p = args, f = files; i < n; i++, p++) {
        isdir[i] = 0;
        if (!File_make(f, ls_flags, pfx, npfx, *p, strlen(*p))) {
            path_last = NULL;
	    continue;
        }
        if (isDirectory(f, NULL)) {
            isdir[i] = 1;
	    path_last = NULL;		/* force path recopy on File re-use */
            if (flags & CLS_NODIR)
                continue;
        }
        if (!(flags & CLS_EXPLICIT) &&	/* skip if in-exclude | not in-include*/
            ((nexclude &&  int_in(f->key,  exclude, nexclude)) ||
	     (nExclude &&  int_in(f->type, Exclude, nExclude)) ||
	     (ninclude && !int_in(f->key,  include, ninclude)) ||
	     (nInclude && !int_in(f->type, Include, nInclude))))
	    path_last = NULL;		/* force path recopy on File re-use */
	else
	    f++;
    }
    nfiles = f - files;
    list = (pFile *) malloc(nfiles * sizeof(pFile));
    for (i = 0; i < nfiles; i++)
	list[i] = files + i;
    if (cmps.n > 0)
        File_sort(list, nfiles);
    strs = (char **) malloc((nfiles + 1) * sizeof *strs);
    lens = (int   *) malloc((nfiles + 1) * sizeof *lens);
    format(list, nfiles, strs, lens);
    ws = tab_layout(lens, nfiles, width, gap, mxcols, mxpad, &nrow, &ncol);
    tab_write(stdout, strs, lens, nfiles, ws, nrow, ncol, EMPTY);
    free((void *) ws);
    free((void *) list);
    for (i = 0; i < nfiles; i++) {
	if (files[i].link > (File *)1) {
	    free(files[i].link->name);
	    free(files[i].link);
	}
        free(strs[i]);
    }
    free(lens);
    free(strs);
    free((void *) files);
    path_last = NULL;   /* force path() recopy in case File* addrs get reused */
}

char **getdir(char *dpath, int *nEnt) { /* return allocated array of strings */
    DIR            *dp;
    struct dirent  *de;
    struct stat     st;
    int             nent = 0, nentA = 8, n, dfd;
    char          **ents, *p;

    if (!(dp = opendir(dpath))) {
        fprintf(stderr, "cannot open \"%s\": %s\n", dpath, strerror(errno));
        return NULL;
    }
    dfd = dirfd(dp);
    fstat(dfd, &st);
    if (!itab_set(&dirtab, st.st_dev, st.st_ino)) {
        closedir(dp);
        return NULL;
    }
    ents = (char **) malloc(nentA * sizeof *ents);
    while ((de = readdir(dp))) {
        p = de->d_name;
        if ((p[0] == '.' && !p[1]) || (p[0] == '.' && p[1] == '.' && !p[2]))
            continue;
        if (nent == nentA) {
            nentA *= 2;
            ents = (char **) realloc(ents, nentA * sizeof *ents);
        }
#if defined __FreeBSD__ || defined __OpenBSD__
	n = de->d_namlen + 1;
#else
	n = strlen(de->d_name) + 1;
#endif
        ents[nent] = (char *)malloc(n);
        memcpy(ents[nent], de->d_name, n);
        nent++;                         /*XXX save 'n' for later somehow */
    }
    ents = realloc(ents, (++nent) * sizeof *ents);
    ents[nent - 1] = NULL;
    *nEnt = nent - 1;
    if (closedir(dp) == -1)
        fprintf(stderr, "cannot close \"%s\": %s\n", dpath, strerror(errno));
    return ents;
}

void lsdir(char *dpath, int r) {      /*** labeling+recursing ls()-wrapper */
    int     nent  = 0, i;
    char  **ents  = getdir(dpath, &nent),
           *isdir = (char *)malloc(nent);
    if (!ents)
        return;
    if (dirlabel) {
        if (dpath[0] == '.' && dpath[1] == '/')     /* XXX loop&remove more junk */
            dpath += 2;
        printf("%s%s%s%s:\n", nl, Color(Types[typeDir].attrON), dpath, attrOFF);
    }
    dirlabel = 1;                   /** Always need dirlabels+line sep after */
    *nl = '\n';                     /** first time we *may* have needed one. */
    ls(dpath, ents, nent, 0, isdir);
    for (i = 0; i < nent; i++)       /* release early...before recursion */
        if (!isdir[i])
            free((void *)ents[i]);   /* could compact & sort in place now...*/
    for (i = 0; i < nent; i++) {
        char subdir[1024];
        if (!isdir[i])
            continue;
        if (r && strcmp(ents[i], ".") != 0 && strcmp(ents[i], "..") != 0) {
            sprintf(subdir, "%s/%s", dpath, ents[i]); /* Q: '/' redundant? */
            lsdir(subdir, r - 1);
        }
        free((void *)ents[i]);
    }
    free((void *)isdir);
    free((void *)ents);
}

char cfg_default[] =
#include "clsrc.h"
;
int cfg_init() {
    char buf[16384], *cp = buf, hmcfg[1024], *tmp;
    int  n, fd;

    if (!cfg && (tmp = getenv("CLS_CONFIG")) && access(tmp, R_OK) == 0)
        cfg  = tmp;
    if (!cfg && getenv("HOME")) {
        sprintf(hmcfg, "%s/.clsrc", getenv("HOME"));
        if (access(hmcfg, R_OK) == 0)
	    cfg = hmcfg;
    }
    if (!cfg && access(SYS_CLSRC, R_OK) == 0)
	cfg = SYS_CLSRC;
    if (cfg) {
        if ((fd = open(cfg, O_RDONLY)) < 0 && (fd = open(cfg, O_RDONLY)) < 0) {
	    if (!quiet)
		fprintf(stderr, "cannot read config file \"%s\": %s\n", cfg,
		        strerror(errno));
            cp = cfg_default;
	} else {
            if ((n = read(fd, buf, sizeof buf - 1)) < 0)
                fprintf(stderr, "read() on config file %s: %s\n", cfg,
			strerror(errno));
            close(fd);
	    buf[n] = '\0';
        }
    } else
	cp = cfg_default;
    if (parse_config(cp))
        return 3;
    return 0;
}

char const * const usage =
"Usage: cls OPTIONS [ FILES... ]\n"
" -v print cls version and exit\n"
" -c CFG_FILE (overrides of CLS_CONFIG, ~/.clsrc, /etc/clsrc, builtin)\n"
" -f FMT_SPEC  (e.g.: -f \"%P %-2.2U %s %m %f%R\")\n"
"    FMT_SPEC has printf-syntax with specifiers converted to %s for output:\n"
"      u user id         a access time   d minor devno    f file name\n"
"      U user name       m modify time   D major devno    F base name\n"
"      g group id        c create time   n link count     r readlink\n"
"      G group name      M max(ctm,mtm)  s size           R lnk w/color tgt\n"
"      t type sort key   i i-number      p permission     S rnd size|devno\n"
"      T file typeno     b file blocks   k ls kind-code   P color perms\n"
" -o ORDER_SPEC  (e.g.: -o tTeN)\n"
"    ORDER_SPEC = [-]KEY1[-]KEY2... ('-' => reverse order.  KEYn may be:\n"
"      u user id         a access time   d minor dev num  f file name\n"
"      U user name       m modify time   D major dev num  F base name\n"
"      g group id        c create time   n link count     e shortest extension\n"
"      G group name      M max(ctm,mtm)  s size           E longest extension\n"
"      t type sort key   i i-number      p permission     L filename length\n"
"      T file typeno     b file blocks                    N numeric file name\n"
"  -t FMT_TM\n"
"     FMT_TM is any valid strftime(3) format.  See local man page for details.\n"
"  -d           list only directories, not contents\n"
"  -l           dereference symlink targets for target-based sorts/formats\n"
"  -L           dereference directory entries which are symlinks\n"
"  -A           use access(2) for perms rather than trusting mode bits\n"
"  -p           text embellishment->off (default=on)\n"
"  -x t1,t2,..  eXclude types t1, t2, ... (as named in config file)\n"
"  -i t1,t2,..  include types t1, t2, ...\n"
"  -X t1,t2,..  EXclude format types t1, t2, ...\n"
"  -I t1,t2,..  Include format types t1, t2, ...\n"
"  -s STYLE     insert command line arguments for STYLE (from config file)\n"
"  -r N         recurse N levels, 0=infinite\n"
"  -<N> (1..9)  use no more than N columns (e.g. -1 means single column)\n";

#include "getopt_r.c"

int parse_cl(int ac, char **av, int *justdir, int *recurse) {
    static int cleared_age; /*NOTE This function is called up to 3 times and */
    int i, ch;              /*     recursively making toggles hard to code. */
    struct getopt_r opt;    /*     So, caller saves flags, mucks w/quiet. */

    getopt_r_init(&opt, "c:f:o:t:dlLApP123456789x:X:i:I:r:s:v");
    while ((ch = getopt_r(&opt, ac, av)) != -1) switch (ch) {
        case 'c': cfg  = opt.arg;                                 break;
        case 'f': fmt  = opt.arg;                                 break;
        case 'o': ord  = opt.arg;                                 break;
        case 't': if (!cleared_age) { cleared_age = 1; nage = 0; }
		  {
		      char *tmp;
		      age[nage] = strtol(opt.arg, &tmp, 10);
		      tfmt[nage] = tmp > opt.arg ? tmp + 1 : tmp;
		      nage++;
		  }
		  break;
        case 'd': *justdir = 1;                                   break;
        case 'l': ls_flags ^= CLS_DEREF_TGT;                      break;
        case 'L': ls_flags ^= CLS_DEREF;                          break;
        case 'A': use_access = 1;                                 break;
        case 'p': use_color = 0;                                  break;
        case '1': case '2': case '3': case '4': case '5': case '6': case '7':
        case '8': case '9': mxcols = ch - '0';                    break;
        case 'x': parse_typelist(opt.arg, exclude, &nexclude, 0); break;
        case 'X': parse_typelist(opt.arg, Exclude, &nExclude, 1); break;
        case 'i': parse_typelist(opt.arg, include, &ninclude, 0); break;
        case 'I': parse_typelist(opt.arg, Include, &nInclude, 1); break;
        case 'r': *recurse = -(atoi(opt.arg) + 1);                break;
        case 's': if ((i = find_style(opt.arg)))
                        parse_cl(n4style[i], v4style[i], justdir, recurse);
                  else if (!quiet)
                        fprintf(stderr, "unknown style \"%s\"\n", opt.arg);
                  break;
        case 'v': printf("%s\n", version);
                  exit(0);
        case -3:  printf("Option \'%c\' requires an argument\n", opt.opt);
                  fputs(usage, stderr);
                  exit(1);
        case -2:  printf("Bad option \'%c\'\n", opt.opt);
                  fputs(usage, stderr);
                  exit(1);
        case -1:  return opt.ind;
        default: fputs(usage, stderr);
                    exit(1);
    }
    return opt.ind;
}

int main(int ac, char **av) {
    int   justdir = 0, recurse  = 0, n, i;
    char *av_def[2] = { "." }, varname[] = "CLS_TIME_FMT_X", *tmp, *tmp2;

    for (i = 0;   i < 256; i++) TOLOWER(i) = i;
    for (i = 'A'; i < 'Z'; i++) TOLOWER(i) = i - 'A' + 'a';
    if ((tmp = getenv("CLS_FORMAT")))   fmt  = tmp;
    if ((tmp = getenv("CLS_ORDER")))    ord  = tmp;
    if ((tmp = getenv("CLS_NOCOLOR")))  use_color = 0;
    now = time(NULL);
    for (i = 0; i < sizeof age / sizeof *age; i++) {
	varname[13] = '0' + i;                  /* 13 = len("CLS_TIME_FMT_") */
	if ((tmp = getenv(varname))) {
	    age[i] = strtol(tmp, &tmp2, 10);
	    if (tmp2 == tmp) {
		age[i] = -2147483648;
		tfmt[i] = tmp;
		nage = i + 1;
		break;
	    } else
	    tfmt[i] = tmp2 + 1;
	    nage = i + 1;
	}
    }
    dirtab = itab_new(NULL, 4);
    gap    = 1;                                 /* min inter-column gap */
    mxcols = 999;                               /* max number of columns */
    if (getenv("CLS_WIDTH"))
        width = atoi(getenv("CLS_WIDTH"));
    if (!width) width  = tab_term_cols();
    init_usr_grp();
    quiet = 1;                                  /* parse quietly 1st time */
    parse_cl(ac, av, &justdir, &recurse);       /* just for -c, etc. */
    ls_flags = 0;                               /* block update by 1st call */
    quiet = 0;                                  /* print errors 2nd time */
    if (cfg_init())				/* use -c, etc. */
        return 4;
    n = parse_cl(ac, av, &justdir, &recurse);
    ac -= n; av += n;				/* shift option args away */
    if ((i = find_style("notty")) && !isatty(1))
	parse_cl(n4style[i], v4style[i], &justdir, &recurse);
    if (!use_color)
        *attrOFF = '\0';
    if (recurse < -1)
        recurse = -1 - recurse;
    ls_flags &= ~CLS_USER | ~CLS_GROUP;		/* let fmt parsers set bits */
    if (parse_fmtspec(fmt))
        return 5;
    if (parse_cmpspec(ord))
        return 6;
    if (nexclude || ninclude || strchr(ord, 't'))
        ls_flags |= CLS_KEY;
    if (nExclude || nInclude || strchr(ord, 'T') || use_color)
        ls_flags |= CLS_TYPE;
    if (ls_flags & CLS_DEREF)			/* -L overrides -l */
	ls_flags &= ~CLS_DEREF_TGT;
    if (((ls_flags & CLS_NEED_TGT) && use_color) ||
        ((ls_flags & CLS_DEREF_TGT) && strchr(ord, 't')))
            ls_flags |= CLS_NEED_TGT;
    if ((ls_flags & CLS_DEREF_TGT) && strchr(ord, 'T'))
        ls_flags |= CLS_TYPE_TGT;
    if (ac <= 0) {
        ac = 1;
        av = av_def;
    }
    if (justdir)                                /* NOTE: => no recursion */
        ls(EMPTY, av, ac, CLS_EXPLICIT, malloc(ac+1));
    else {
        int   i, nd = 0;
        char *isdir = (char *)malloc(ac);

        ls(EMPTY, av, ac, CLS_NODIR | CLS_EXPLICIT, isdir);
        *nl = '\n';
        dirlabel = 1;                           /* dirlabel in lsdir() */
        for (i = 0; i < ac; i++)                /* count directories */
            if (isdir[i])
                nd++;
        if (nd == ac) {                        /** all av are directories */
            *nl = '\0';                         /* No 1st \n if av all dirs */
            if (!recurse && nd == 1)
                dirlabel = 0;                   /* no label when av===1 dir */
        }
        for (i = 0; i < ac; i++)
            if (isdir[i])
                lsdir(av[i], recurse);
    }
    return exit_code;
}
