#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

typedef struct {
    unsigned	link;
    dev_t	dev;
    ino_t	ino;
} itab_slot;
#define ISLOT_VACANT(slot)	  ((slot)->dev == 0)
#define ISLOT_EQ(slot, _d_, _i_)  ((slot)->dev==(_d_) && (slot)->ino==(_i_))
#define IHASH(_d_,_i_)            ((unsigned)(0x4c6d92b3U * ino + dev))

typedef struct {
    unsigned	shift : 5;
    unsigned	pop   : 27;                     /* up to 134 million entries */
    unsigned    r;
    itab_slot   tab[1];
} itab_t;
#define ITAB_SZ(_t_)        (1 << (32 - (_t_)->shift))
#define ITAB_NIL            ((unsigned) -1)

itab_t *itab_new(itab_t *t, unsigned po2) {
    if (!t) t = malloc(sizeof *t + ((1 << po2) - 1) * sizeof(itab_slot));
    if (!t) return NULL;
    memset(t, 0, sizeof *t + ((1 << po2) - 1) * sizeof(itab_slot));
    t->r     = 1 << po2;
    t->shift = 32 - po2;
    t->pop   = 0;
    return t;
}

int itab_resize(itab_t **T, unsigned new_po2);

int itab_set(itab_t **T, dev_t dev, ino_t ino) { /* 1 on good put of NOVEL key*/
    unsigned   h    = IHASH(dev, ino);          /* compute hash */
    itab_slot *slot = (*T)->tab + (h >> (*T)->shift);
    itab_t    *t    = *T;

    if (t->pop << 3 > 7 * ITAB_SZ(t) && !itab_resize(T, 2+(32 - t->shift)))
        return 0;                               /* out of memory */
    t = *T;                                     /* in case resize relocated */
    slot = t->tab + (h >> t->shift);            /* first table probe */
    if (ISLOT_VACANT(slot))                     /* common two cases first */
        goto ln;
lp: if (ISLOT_EQ(slot, dev, ino))               /* already present */
        return 0;
    if (slot->link != ITAB_NIL) {               /* search collision chain */
        slot = t->tab + slot->link;
        goto lp;
    }
    while (--t->r && !ISLOT_VACANT(&t->tab[t->r]))
            /**/;                               /* alloc non-init-probe slot */
    slot->link = t->r;                          /* last in chain->new slot */
    slot = t->tab + t->r;                       /* new slot=newly allocd t->r */
ln: slot->link = ITAB_NIL;                      /* terminate collision chain */
    slot->dev = dev;                            /* store key */
    slot->ino = ino;
    t->pop++;                                   /* update population counter */
    return 1;
}

int itab_resize(itab_t **T, unsigned new_po2) {
    itab_t    *t_old = *T, *t_new = itab_new(NULL, new_po2);
    itab_slot *slot, *end = t_old->tab + ITAB_SZ(t_old);

    if (!t_new)                                 /* malloc failed: fail */
        return write(2, "out of memory\n", 14), 0;
    for (slot = t_old->tab; slot < end; slot++)
        if (!ISLOT_VACANT(slot))
            itab_set(&t_new, slot->dev, slot->ino);
    *T = t_new;
    free(t_old);
    return 1;
}
