/* Bentley,McIlroy: "Engineering a Sort Function".  #include'r must define:  */
/*   S_() S_IDX_T S_VAL_T S_LESS(x,y) S_LESSEQ(x,y) S_EQ(x,y) S_GREATER(x,y) */
/* NOTE: comparator macros take ADDRESSES of data, not values themselves.    */

/* #define UNROLL_POW 4 */
#ifndef S_
#define S_(x) S_##x
#endif

#ifdef S_CMP		/* Allows optimization to single comparison */
#ifndef S_LESS
#define S_LESS(a,b)	(S_CMP(a,b) < 0)
#endif
#ifndef S_LESSEQ
#define S_LESSEQ(a,b)	(S_CMP(a,b) <= 0)
#endif
#ifndef S_EQ
#define S_EQ(a,b)	(S_CMP(a,b) == 0)
#endif
#ifndef S_GREATER
#define S_GREATER(a,b)	(S_CMP(a,b) > 0)
#endif
#endif

#ifdef __STRICT_ANSI__
#define inline __inline
#endif

static inline S_VAL_T *S_(med3) (S_VAL_T *a, S_VAL_T *b, S_VAL_T *c) {
	return S_LESS(a,b) ? (S_LESS   (b,c) ? b : (S_LESS(a,c) ? c : a ))
                           : (S_GREATER(b,c) ? b : (S_LESS(a,c) ? a : c ));
}

static inline void S_(swap) (S_VAL_T *i, S_VAL_T *j) {
	S_VAL_T t = *i;
	*i = *j;
	*j = t;
}

static inline void S_(move) (S_VAL_T *i, S_VAL_T *j, S_IDX_T n, S_IDX_T m) {
	S_VAL_T T;

	if (m < n)
		n = m;
	j = j - n;

#ifdef UNROLL_POW	/* This speeds things up only slightly for doubles */
#define BODY  (T = *i, *i++ = *j, *j++ = T)
	m = n >> UNROLL_POW;
	switch (n & ((1 << UNROLL_POW) - 1)) {
		case  0: while (m-- > 0) { BODY;
#if UNROLL_POW >= 4
		case 15:                   BODY;
		case 14:                   BODY;
		case 13:                   BODY;
		case 12:                   BODY;
		case 11:                   BODY;
		case 10:                   BODY;
		case  9:                   BODY;
		case  8:                   BODY;
#endif
#if UNROLL_POW >= 3
		case  7:                   BODY;
		case  6:                   BODY;
		case  5:                   BODY;
		case  4:                   BODY;
#endif
#if UNROLL_POW >= 2
		case  3:                   BODY;
		case  2:                   BODY;
#endif
		case  1:                   BODY;
			 }
	}
#undef BODY
#else
	while (n > 0) {		/* unroll this loop? */
		T    = *i;
		*i++ = *j;
		*j++ = T;
		n--;
        }
#endif
}

static void S_SORT (S_VAL_T *X, S_IDX_T N) {
	S_VAL_T *a, *b, *c, *d,  *l, *m, *n,  *x = X;
	int	 r, swap_cnt;
loop:
	swap_cnt = 0;
	if (N < 7) {
		for (m = x + 1; m < x + N; m++)
			for (l = m; l > x && !S_LESSEQ (l - 1, l); l--)
				S_(swap) (l, l - 1);
		return;
	}
	m = x + (N >> 1);
	if (N > 7) {
		l = x;
		n = x + (N - 1);
		if (N > 40) {
			S_IDX_T D = N >> 3;
			l = S_(med3) (l, l + D, l + 2 * D);
			m = S_(med3) (m - D, m, m + D);
			n = S_(med3) (n - 2 * D, n - D, n);
		}
		m = S_(med3) (l, m, n);
	}
	S_(swap) (x, m);
	a = b = x + 1;
	c = d = x + (N - 1);
	for (;;) {
		while (b <= c && S_LESSEQ (b, x)) {
			if (S_EQ (x, b)) {
				swap_cnt = 1;
				S_(swap) (a, b);
				a++;
			}
			b++;
		}
		while (b <= c && S_LESS (x, c)) {
			if (S_EQ (x, c)) {
				swap_cnt = 1;
				S_(swap) (c, d);
				d --;
			}
			c --;
		}
		if (b > c)
			break;
		S_(swap) (b, c);
		swap_cnt = 1;
		b++;
		c--;
	}
	if (swap_cnt == 0) {		/* Switch to insertion sort */
		for (m = x + 1; m < x + N; m++)
			for (l = m; l > x && S_GREATER (l - 1, l); l--)
				S_(swap) (l, l - 1);
		return;
	}
	n = x + N;
	S_(move) (x, b, a - x, b - a);
	S_(move) (b, n, d - c, n - d - 1);
	if ((r = b - a) > 1)
		S_SORT (x, r);
	if ((r = d - c) > 1) { 
		x = n - r;	/* Iterate rather than recurse */
		N = r;
		goto loop;
	}
}
#ifndef S_PRESERVE
# undef S_IDX_T
# undef S_VAL_T
# undef S_LESS
# undef S_LESSEQ
# undef S_EQ
# undef S_GREATER
# undef S_CMP
# undef S_SORT
# undef S_
#endif /* S_PRESERVE */
